@@ -6,11 +6,7 @@
 pref("startup.homepage_override_url","http://www.palemoon.org/releasenotes.shtml");
 pref("app.releaseNotesURL", "http://www.palemoon.org/releasenotes.shtml");
 
-// Enable Firefox compatmode by default.
-pref("general.useragent.compatMode", 2);
-pref("general.useragent.compatMode.gecko", true);
-pref("general.useragent.compatMode.firefox", true);
-pref("general.useragent.compatMode.version", "102.0");
+// Firefox compatmode is set in shared prefs.
 
 // Enable dynamic UA updates
 pref("general.useragent.updates.enabled", true);
