@@ -34,6 +34,10 @@ class ObjectBox;
     F(POSTDECREMENT) \
     F(DOT) \
     F(ELEM) \
+    F(OPTDOT) \
+    F(OPTCHAIN) \
+    F(OPTELEM) \
+    F(OPTCALL) \
     F(ARRAY) \
     F(ELISION) \
     F(STATEMENTLIST) \
@@ -76,6 +80,7 @@ class ObjectBox;
     F(DELETEPROP) \
     F(DELETEELEM) \
     F(DELETEEXPR) \
+    F(DELETEOPTCHAIN) \
     F(TRY) \
     F(CATCH) \
     F(CATCHLIST) \
@@ -125,8 +130,12 @@ class ObjectBox;
     \
     /* \
      * Binary operators. \
-     * These must be in the same order as TOK_OR and friends in TokenStream.h. \
+     * These must be in the same order in several places: \
+     *   - the precedence table and JSOp code list in Parser.cpp \
+     *   - the binary operators in TokenKind.h \
+     *   - the first and last binary operator markers in ParseNode.h \
      */ \
+    F(COALESCE) \
     F(OR) \
     F(AND) \
     F(BITOR) \
@@ -184,7 +193,13 @@ enum ParseNodeKind
     FOR_EACH_PARSE_NODE_KIND(EMIT_ENUM)
 #undef EMIT_ENUM
     PNK_LIMIT, /* domain size */
-    PNK_BINOP_FIRST = PNK_OR,
+    /*
+     * Binary operator markers.
+     * These must be in the same order in several places:
+     *   - the precedence table and JSOp code list in Parser.cpp
+     *   - the binary operators in ParseNode.h and TokenKind.h
+     */
+    PNK_BINOP_FIRST = PNK_COALESCE,
     PNK_BINOP_LAST = PNK_POW,
     PNK_ASSIGNMENT_START = PNK_ASSIGN,
     PNK_ASSIGNMENT_LAST = PNK_POWASSIGN
@@ -324,8 +339,9 @@ IsTypeofKind(ParseNodeKind kind)
  * PNK_POWASSIGN
  * PNK_CONDITIONAL ternary  (cond ? trueExpr : falseExpr)
  *                          pn_kid1: cond, pn_kid2: then, pn_kid3: else
- * PNK_OR,      list        pn_head; list of pn_count subexpressions
- * PNK_AND,                 All of these operators are left-associative except (**).
+ * PNK_COALESCE,   list     pn_head; list of pn_count subexpressions
+ * PNK_OR,                  All of these operators are left-associative except (**).
+ * PNK_AND
  * PNK_BITOR,
  * PNK_BITXOR,
  * PNK_BITAND,
@@ -371,6 +387,29 @@ IsTypeofKind(ParseNodeKind kind)
  *                          for a more-specific PNK_DELETE* unless constant
  *                          folding (or a similar parse tree manipulation) has
  *                          occurred
+ * PNK_DELETEOPTCHAIN unary pn_kid: MEMBER expr that's evaluated, then the
+ *                          overall delete evaluates to true; If constant
+ *                          folding occurs, PNK_ELEM may become PNK_DOT.
+ *                          PNK_OPTELEM does not get folded into PNK_OPTDOT.
+ * PNK_OPTCHAIN unary       pn_kid: expression that is evaluated as a chain.
+ *                          Contains one or more optional nodes. Its first node
+ *                          is always an optional node, such as PNK_OPTELEM,
+ *                          PNK_OPTDOT, or PNK_OPTCALL. This will shortcircuit
+ *                          and return 'undefined' without evaluating the rest
+ *                          of the expression if any of the optional nodes it
+ *                          contains are nullish. An optional chain can also
+ *                          contain nodes such as PNK_DOT, PNK_ELEM, PNK_NAME,
+ *                          PNK_CALL, etc. These are evaluated normally.
+ * PNK_OPTDOT   name        pn_expr: MEMBER expr to left of .
+ *                          short circuits back to PNK_OPTCHAIN if nullish.
+ *                          pn_atom: name to right of .
+ * PNK_OPTELEM  binary      pn_left: MEMBER expr to left of [
+ *                          short circuits back to PNK_OPTCHAIN if nullish.
+ *                          pn_right: expr between [ and ]
+ * PNK_OPTCALL  list        pn_head: list of call, arg1, arg2, ... argN
+ *                          pn_count: 1 + N (where N is number of args)
+ *                          call is a MEMBER expr naming a callable object,
+ *                          short circuits back to PNK_OPTCHAIN if nullish.
  * PNK_DOT      name        pn_expr: MEMBER expr to left of .
  *                          pn_atom: name to right of .
  * PNK_ELEM     binary      pn_left: MEMBER expr to left of [
@@ -1182,11 +1221,12 @@ class RegExpLiteral : public NullaryNode
     }
 };
 
-class PropertyAccess : public ParseNode
+class PropertyAccessBase : public ParseNode
 {
   public:
-    PropertyAccess(ParseNode* lhs, PropertyName* name, uint32_t begin, uint32_t end)
-      : ParseNode(PNK_DOT, JSOP_NOP, PN_NAME, TokenPos(begin, end))
+    PropertyAccessBase(ParseNodeKind kind, ParseNode* lhs, PropertyName* name,
+                       uint32_t begin, uint32_t end)
+      : ParseNode(kind, JSOP_NOP, PN_NAME, TokenPos(begin, end))
     {
         MOZ_ASSERT(lhs != nullptr);
         MOZ_ASSERT(name != nullptr);
@@ -1195,7 +1235,8 @@ class PropertyAccess : public ParseNode
     }
 
     static bool test(const ParseNode& node) {
-        bool match = node.isKind(PNK_DOT);
+        bool match = node.isKind(PNK_DOT) ||
+                     node.isKind(PNK_OPTDOT);
         MOZ_ASSERT_IF(match, node.isArity(PN_NAME));
         return match;
     }
@@ -1207,6 +1248,24 @@ class PropertyAccess : public ParseNode
     PropertyName& name() const {
         return *pn_u.name.atom->asPropertyName();
     }
+};
+
+class PropertyAccess : public PropertyAccessBase
+{
+  public:
+    PropertyAccess(ParseNode* lhs, PropertyName* name, uint32_t begin,
+                   uint32_t end)
+      : PropertyAccessBase(PNK_DOT, lhs, name, begin, end)
+    {
+        MOZ_ASSERT(lhs != nullptr);
+        MOZ_ASSERT(name != nullptr);
+    }
+
+    static bool test(const ParseNode& node) {
+        bool match = node.isKind(PNK_DOT);
+        MOZ_ASSERT_IF(match, node.isArity(PN_NAME));
+        return match;
+    }
 
     bool isSuper() const {
         // PNK_SUPERBASE cannot result from any expression syntax.
@@ -1214,16 +1273,49 @@ class PropertyAccess : public ParseNode
     }
 };
 
-class PropertyByValue : public ParseNode
+class OptionalPropertyAccess : public PropertyAccessBase
 {
   public:
-    PropertyByValue(ParseNode* lhs, ParseNode* propExpr, uint32_t begin, uint32_t end)
-      : ParseNode(PNK_ELEM, JSOP_NOP, PN_BINARY, TokenPos(begin, end))
+    OptionalPropertyAccess(ParseNode* lhs, PropertyName* name, uint32_t begin,
+                           uint32_t end)
+      : PropertyAccessBase(PNK_OPTDOT, lhs, name, begin, end)
+    {
+        MOZ_ASSERT(lhs != nullptr);
+        MOZ_ASSERT(name != nullptr);
+    }
+
+    static bool test(const ParseNode& node) {
+        bool match = node.isKind(PNK_OPTDOT);
+        MOZ_ASSERT_IF(match, node.isArity(PN_NAME));
+        return match;
+    }
+};
+
+class PropertyByValueBase : public ParseNode
+{
+  public:
+    PropertyByValueBase(ParseNodeKind kind, ParseNode* lhs, ParseNode* propExpr,
+                    uint32_t begin, uint32_t end)
+      : ParseNode(kind, JSOP_NOP, PN_BINARY, TokenPos(begin, end))
     {
         pn_u.binary.left = lhs;
         pn_u.binary.right = propExpr;
     }
 
+    static bool test(const ParseNode& node) {
+        bool match = node.isKind(PNK_ELEM) ||
+                     node.isKind(PNK_OPTELEM);
+        MOZ_ASSERT_IF(match, node.isArity(PN_BINARY));
+        return match;
+    }
+};
+
+class PropertyByValue : public PropertyByValueBase {
+  public:
+    PropertyByValue(ParseNode* lhs, ParseNode* propExpr, uint32_t begin,
+                    uint32_t end)
+      : PropertyByValueBase(PNK_ELEM, lhs, propExpr, begin, end) {}
+
     static bool test(const ParseNode& node) {
         bool match = node.isKind(PNK_ELEM);
         MOZ_ASSERT_IF(match, node.isArity(PN_BINARY));
@@ -1235,6 +1327,19 @@ class PropertyByValue : public ParseNode
     }
 };
 
+class OptionalPropertyByValue : public PropertyByValueBase {
+ public:
+    OptionalPropertyByValue(ParseNode* lhs, ParseNode* propExpr,
+                            uint32_t begin, uint32_t end)
+      : PropertyByValueBase(PNK_OPTELEM, lhs, propExpr, begin, end) {}
+
+    static bool test(const ParseNode& node) {
+        bool match = node.isKind(PNK_OPTELEM);
+        MOZ_ASSERT_IF(match, node.isArity(PN_BINARY));
+        return match;
+    }
+};
+
 /*
  * A CallSiteNode represents the implicit call site object argument in a TaggedTemplate.
  */
