@@ -80,6 +80,8 @@ public:
   // the first <label> element is clicked, that one may set this true.
   // Then, the second <label> element won't handle the event.
   bool    mMultipleActionsPrevented : 1;
+  // Similar to above but expected to be used during PreHandleEvent phase.
+  bool    mMultiplePreActionsPrevented : 1;
   // If mIsBeingDispatched is true, the DOM event created from the event is
   // dispatching into the DOM tree and not completed.
   bool    mIsBeingDispatched : 1;
